/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>

#define PN_MIDI_IN 0
#define PN_MIDI_OUT 1
#define PN_CHANNEL 2
#define PN_PROG_IN 3
#define PN_PROG_OUT 4


#define SYSTEM_MSG		        0xF0
#define PROG_CHANGE             0xC0

// String used for channel
char chanstr[512];
char proginstr[2048];
char progoutstr[2048];

void BuildStrings()
    {
    static bool need_build = true;

    if(need_build)
        {
        char tmp[40];
        int x;

        // Build the channel string
        strcpy(chanstr, "Any=-1");

        for(x = 0; x < 16; x++)
            {
            sprintf(tmp, ",%d=%d", x + 1, x);
            strcat(chanstr, tmp);
            }

        // Build the program input string
        strcpy(proginstr, "None=-2,Any=-1");
        
        for(x = 0; x < 128; x++)
            {
            sprintf(tmp, ",%d=%d", x + 1, x);
            strcat(proginstr, tmp);
            }

        // Build the program output string
        strcpy(progoutstr, "1=0");

        for(x = 1; x < 128; x++)
            {
            sprintf(tmp, ",%d=%d", x + 1, x);
            strcat(progoutstr, tmp);
            }
        }


    need_build = false;
    }



// The core of a MIDI plugin
void Module::OnMidiData( unsigned long p_clock, unsigned long midi_msg, short pin_id)
{
	// sytem messages are for timing info etc
	// they don't have a channel, just pass them right through
	bool is_system_msg = ( midi_msg & SYSTEM_MSG ) == SYSTEM_MSG;

	if( !is_system_msg )
	{
		// The first byte of a midi message contains the status (note on/off etc) and the channel
		int chan = midi_msg & 0x0f;
		int stat = midi_msg & 0xf0;

		// the next two bytes depend on the type of message
		int b2 = ( midi_msg / 0x100 ) & 0xff;
		int b3 = ( midi_msg / 0x10000 ) & 0xff;

		switch( stat )
		{
		case PROG_CHANGE:
			{
            // Remember original program
            original[chan] = b2;

            if((channel == chan || channel == -1) && (prgin == b2 || prgin == -1))
                {
                b2 = prgout;
                }

            break;   
			}

		}

    	// re-assemble the MIDI message into an 32 bit integer
    	midi_msg = stat + chan + (b2 << 8) + (b3 << 16);

	}

	// send the MIDI
	getPin(PN_MIDI_OUT)->TransmitMIDI( p_clock, midi_msg );
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
int x;
for(x = 0; x < 16; x++)
    {
    original[x] = 0;
    }

prgin = prgout = channel = 0;
}

void Module::open()
{
	SEModule_base::open();

	// choose which function is used to process audio
	SET_PROCESS_FUNC(Module::sub_process);
       
}

Module::~Module()
{
	// nothing to see here, move along.
}

// describe the module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	properties->name = "MIDI Program Changer";	// this is the name the end-user will see

	// return a unique string.
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug. No two plugs may have the same id.
	properties->id = "BAVII MIDIProgChanger";

	// Info, may include Author, Web page whatever
	properties->about = "by Brian Vanderburg II";
	return true;
}

// describe the pins (plugs)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
    BuildStrings();

	switch( index )
	{
	case PN_MIDI_IN:
		properties->name				= "MIDI In";
		properties->direction			= DR_IN;
		properties->datatype			= DT_MIDI2;
		break;
    case PN_MIDI_OUT:
		properties->name				= "MIDI Out";
		properties->direction			= DR_OUT;
		properties->datatype			= DT_MIDI2;
		break;
    case PN_CHANNEL:
		properties->name				= "Channel";
		properties->variable_address	= &channel;
		properties->datatype_extra		= chanstr;
		properties->default_value		= "-1";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	case PN_PROG_IN:
		properties->name				= "Program In";
		properties->variable_address	= &prgin;
		properties->datatype_extra		= proginstr;
		properties->default_value		= "0";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
    case PN_PROG_OUT:
		properties->name				= "Program Out";
		properties->variable_address	= &prgout;
		properties->datatype_extra		= progoutstr;
		properties->default_value		= "0";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

void Module::sub_process(long buffer_offset, long sampleFrames )
{
	// Nothing to do here in a pure MIDI plugin
//	if( !AreEventsPending() )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

void Module::OnPlugStateChange(SEPin *pin)
{
    int c;

    unsigned long p_clock = SampleClock();

    

	switch(pin->getPinID())
        {
        case PN_PROG_IN:
        case PN_PROG_OUT:
        case PN_CHANNEL:
            
            // Update the programs for each channel
            for(c = 0; c < 16; c++)
                {
                // If the current channel is active and the program for the
                // channel is the one to modify, then set the modified program
                if((channel == c || channel == -1) && (prgin == original[c] || prgin == -1))
                    {
                    getPin(PN_MIDI_OUT)->TransmitMIDI( p_clock, 0xC0 + c + (prgout << 8) );
                    }
                else
                    {
                    // We do not affect the current channel or program
                    // Set the original program
                    getPin(PN_MIDI_OUT)->TransmitMIDI( p_clock, 0xC0 + c + (original[c] << 8) );
                    }
                }

            break;
        }
}

